==> TundereBird-0.28.tar.bz2: TundereBird-0.28/core/Smdn.Net.Imap4.Client/Smdn.Net.Imap4.Client.Session/ImapSessionCreator.cs <==
// 
// Author:
//       smdn <smdn@mail.invisiblefulmoon.net>
// 
// Copyright (c) 2008-2010 smdn
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Collections.Generic;
using System.Net;

using Smdn.Collections;
using Smdn.Net.Imap4.Client.Transaction;
using Smdn.Net.Imap4.Protocol;
using Smdn.Net.MessageAccessProtocols;
using Smdn.Security.Authentication.Sasl.Client;

namespace Smdn.Net.Imap4.Client.Session {
  public static class ImapSessionCreator {
    public static ImapSession CreateSession(IImapSessionProfile profile,
                                            UpgradeConnectionStreamCallback createSslStreamCallback)
    {
      ImapSession session;

      var result = CreateSession(profile, createSslStreamCallback, out session);

      if (result.Succeeded)
        return session;
      else
        throw new ImapException("create session failed: " + result.ResultText);
    }

    public static ImapTransactionResult CreateSession(IImapSessionProfile profile,
                                                      UpgradeConnectionStreamCallback createSslStreamCallback,
                                                      out ImapSession session)
    {
      if (profile == null)
        throw new ArgumentNullException("profile");

      var authority = profile.Authority;
      var securePort = string.Equals(authority.Scheme, ImapUri.UriSchemeImaps, StringComparison.InvariantCultureIgnoreCase);

      if (securePort && createSslStreamCallback == null)
        throw new ArgumentNullException("createSslStreamCallback");

      ImapTransactionResult result;
      session = null;

      try {
        session = new ImapSession(authority.Host,
                                  authority.Port,
                                  true,
                                  profile.Timeout,
                                  securePort
                                    ? createSslStreamCallback
                                    : null);
      }
      catch (ImapUpgradeConnectionException ex) {
        throw new ImapSecureConnectionException(ex.Message, ex.InnerException);
      }

      session.HandlesIncapableAsException = false;
      session.HandlesReferralAsException = false;
      session.TransactionTimeout  = profile.Timeout;
      session.SendTimeout         = profile.SendTimeout;
      session.ReceiveTimeout      = profile.ReceiveTimeout;

      // try querying server capability (ignore error)
      session.Capability();

      if (!session.ServerCapabilities.Has(ImapCapability.Imap4Rev1))
        throw new ImapIncapableException(ImapCapability.Imap4Rev1);

      if (profile.UseTlsIfAvailable && session.ServerCapabilities.Has(ImapCapability.StartTls) && !session.IsSecureConnection) {
        ImapTransactionResult r = null;

        try {
          r = session.StartTls(createSslStreamCallback, true);
        }
        catch (ImapUpgradeConnectionException ex) {
          throw new ImapSecureConnectionException(ex.Message, ex.InnerException);
        }
        catch (Exception ex) {
          throw ex;
        }

        if (r.Failed)
          throw new ImapSecureConnectionException(r.ResultText);
        else if (!session.ServerCapabilities.Has(ImapCapability.Imap4Rev1))
          throw new ImapIncapableException(ImapCapability.Imap4Rev1);
      }

      if (profile.UseDeflateIfAvailable && session.ServerCapabilities.Has(ImapCapability.CompressDeflate)) {
        var r = session.Compress(ImapCompressionMechanism.Deflate);

        if (r.Failed)
          throw new WebException(r.ResultText, null, WebExceptionStatus.RequestCanceled, null);
      }

      result = Authenticate(session, profile);

      if (result == null) {
        throw new ImapException("appropriate authentication mechanism not found");
      }
      else if (result.Failed) {
        try {
          try {
            session.Disconnect(false);
          }
          catch (ImapConnectionException) {
            // ignore
          }
        }
        finally {
          session = null;
        }
      }

      return result;
    }

    private static ImapTransactionResult Authenticate(ImapSession session, IImapSessionProfile profile)
    {
      var authority = profile.Authority;
      var username = ImapStyleUriParser.GetUser(authority);
      var authMechanism = ImapStyleUriParser.GetAuthType(authority);

      /*
       * http://tools.ietf.org/html/rfc5092
       * 3.2. IMAP User Name and Authentication Mechanism
       * 
       *    An authentication mechanism (as used by the IMAP AUTHENTICATE
       *    command) can be expressed by adding ";AUTH=<enc-auth-type>" to the
       *    end of the user name in an IMAP URL.  When such an <enc-auth-type> is
       *    indicated, the client SHOULD request appropriate credentials from
       *    that mechanism and use the "AUTHENTICATE" command instead of the
       *    "LOGIN" command.  If no user name is specified, one MUST be obtained
       *    from the mechanism or requested from the user/configuration as
       *    appropriate.
       *
       *    If a user name is included with no authentication mechanism, then
       *    ";AUTH=*" is assumed.
       */

      // TODO: URLAUTH
      /*
      if (!session.ServerCapabilities.Has(ImapCapability.UrlAuth)
        throw new ImapWebException("URLAUTH incapable", WebExceptionStatus.ProtocolError);
      */
      var canFallback = false;

      if (authMechanism == null) {
        if (string.IsNullOrEmpty(username)) {
          authMechanism = ImapAuthenticationMechanism.Anonymous;
          canFallback = true;
        }
        else {
          authMechanism = ImapAuthenticationMechanism.SelectAppropriate;
        }
      }

      if (authMechanism == ImapAuthenticationMechanism.SelectAppropriate)
        return AuthenticateWithAppropriateMechanism(session,
                                                    profile.Credentials,
                                                    username,
                                                    GetUsingSaslMechanisms(profile.UsingSaslMechanisms ?? new string[0]));
      else if (authMechanism == ImapAuthenticationMechanism.Anonymous)
        return AuthenticateAsAnonymous(session,
                                       profile.AnonymousCredential,
                                       canFallback);
      else
        return AuthenticateWithSuppliedMechanism(session,
                                                 profile.Credentials,
                                                 username,
                                                 authMechanism);
    }

    private static IEnumerable<ImapAuthenticationMechanism> GetUsingSaslMechanisms(string[] usingSaslMechanisms)
    {
      var availableMechanisms = SaslClientMechanism.GetAvailableMechanisms();

      foreach (var mechanism in usingSaslMechanisms) {
        if (availableMechanisms.Exists(delegate(string val) {
          return string.Equals(val, mechanism, StringComparison.InvariantCultureIgnoreCase);
        }))
          yield return ImapAuthenticationMechanism.GetKnownOrCreate(mechanism);
      }
    }

    private static ImapTransactionResult AuthenticateWithAppropriateMechanism(ImapSession session, ICredentialsByHost credentials, string username, IEnumerable<ImapAuthenticationMechanism> usingSaslMechanisms)
    {
      ImapTransactionResult result = null;

      foreach (var mechanism in usingSaslMechanisms) {
        if (session.ServerCapabilities.IsCapable(mechanism)) {
          result = session.Authenticate(credentials, username, mechanism);

          if (result.Succeeded)
            break;
        }
      }

      if ((result == null || result.Failed) &&
          !session.ServerCapabilities.Has(ImapCapability.LoginDisabled))
        result = session.Login(credentials, username);

      return result;
    }

    private static ImapTransactionResult AuthenticateWithSuppliedMechanism(ImapSession session, ICredentialsByHost credentials, string username, ImapAuthenticationMechanism authMechanism)
    {
      /*
       * http://tools.ietf.org/html/rfc5092
       * 3.2. IMAP User Name and Authentication Mechanism
       * 
       *    An authentication mechanism (as used by the IMAP AUTHENTICATE
       *    command) can be expressed by adding ";AUTH=<enc-auth-type>" to the
       *    end of the user name in an IMAP URL.  When such an <enc-auth-type> is
       *    indicated, the client SHOULD request appropriate credentials from
       *    that mechanism and use the "AUTHENTICATE" command instead of the
       *    "LOGIN" command.  If no user name is specified, one MUST be obtained
       *    from the mechanism or requested from the user/configuration as
       *    appropriate.
       */
      return session.Authenticate(credentials, username, authMechanism);
    }

    private static ImapTransactionResult AuthenticateAsAnonymous(ImapSession session, NetworkCredential anonymousCredential, bool canFallback)
    {
      /*
       * http://tools.ietf.org/html/rfc5092
       * 3.2. IMAP User Name and Authentication Mechanism
       * 
       *    If no user name and no authentication mechanism are supplied, the
       *    client MUST authenticate as anonymous to the server.  If the server
       *    advertises AUTH=ANONYMOUS IMAP capability, the client MUST use the
       *    AUTHENTICATE command with ANONYMOUS [ANONYMOUS] SASL mechanism.  If
       *    SASL ANONYMOUS is not available, the (case-insensitive) user name
       *    "anonymous" is used with the "LOGIN" command and the Internet email
       *    address of the end user accessing the resource is supplied as the
       *    password.  The latter option is given in order to provide for
       *    interoperability with deployed servers.
       * 
       *    Note that, as described in RFC 3501, the "LOGIN" command MUST NOT be
       *    used when the IMAP server advertises the LOGINDISABLED capability.
       */
      ImapTransactionResult result = null;

      if (session.ServerCapabilities.IsCapable(ImapAuthenticationMechanism.Anonymous))
        // try AUTHENTICATE ANONYUMOUS
        result = session.Authenticate(anonymousCredential, ImapAuthenticationMechanism.Anonymous);

      if ((result == null || (result.Failed && canFallback)) &&
          !session.ServerCapabilities.Has(ImapCapability.LoginDisabled))
        // try anonymous LOGIN
        result = session.Login(anonymousCredential);

      return result;
    }
  }
}

==> TundereBird-0.29.tar.bz2: TundereBird-0.29/core/Smdn.Net.Imap4.Client/Smdn.Net.Imap4.Client.Session/ImapSessionCreator.cs <==
// 
// Author:
//       smdn <smdn@mail.invisiblefulmoon.net>
// 
// Copyright (c) 2008-2010 smdn
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Collections.Generic;
using System.Net;

using Smdn.Collections;
using Smdn.Net.Imap4.Client.Transaction;
using Smdn.Net.Imap4.Protocol;
using Smdn.Security.Authentication.Sasl.Client;

namespace Smdn.Net.Imap4.Client.Session {
  public static class ImapSessionCreator {
    public static ImapSession CreateSession(IImapSessionProfile profile,
                                            UpgradeConnectionStreamCallback createSslStreamCallback)
    {
      ImapSession session;

      var result = CreateSession(profile, createSslStreamCallback, out session);

      if (result.Succeeded)
        return session;
      else
        throw new ImapException("create session failed: " + result.ResultText);
    }

    public static ImapTransactionResult CreateSession(IImapSessionProfile profile,
                                                      UpgradeConnectionStreamCallback createSslStreamCallback,
                                                      out ImapSession session)
    {
      if (profile == null)
        throw new ArgumentNullException("profile");

      var authority = profile.Authority;
      var securePort = string.Equals(authority.Scheme, ImapUri.UriSchemeImaps, StringComparison.OrdinalIgnoreCase);

      if (securePort && createSslStreamCallback == null)
        throw new ArgumentNullException("createSslStreamCallback");

      ImapTransactionResult result;
      session = null;

      try {
        session = new ImapSession(authority.Host,
                                  authority.Port,
                                  true,
                                  profile.Timeout,
                                  securePort
                                    ? createSslStreamCallback
                                    : null);
      }
      catch (ImapUpgradeConnectionException ex) {
        throw new ImapSecureConnectionException(ex.Message, ex.InnerException);
      }

      session.HandlesIncapableAsException = false;
      session.HandlesReferralAsException = false;
      session.TransactionTimeout  = profile.Timeout;
      session.SendTimeout         = profile.SendTimeout;
      session.ReceiveTimeout      = profile.ReceiveTimeout;

      if (session.ServerCapabilities.Count == 0)
        // try querying server capability (ignore error)
        session.Capability();

      if (!session.ServerCapabilities.Has(ImapCapability.Imap4Rev1))
        throw new ImapIncapableException(ImapCapability.Imap4Rev1);

      if (profile.UseTlsIfAvailable && session.ServerCapabilities.Has(ImapCapability.StartTls) && !session.IsSecureConnection) {
        ImapTransactionResult r = null;

        try {
          r = session.StartTls(createSslStreamCallback, true);
        }
        catch (ImapUpgradeConnectionException ex) {
          throw new ImapSecureConnectionException(ex.Message, ex.InnerException);
        }
        catch (Exception ex) {
          throw ex;
        }

        if (r.Failed)
          throw new ImapSecureConnectionException(r.ResultText);
        else if (!session.ServerCapabilities.Has(ImapCapability.Imap4Rev1))
          throw new ImapIncapableException(ImapCapability.Imap4Rev1);
      }

      if (profile.UseDeflateIfAvailable && session.ServerCapabilities.Has(ImapCapability.CompressDeflate)) {
        var r = session.Compress(ImapCompressionMechanism.Deflate);

        if (r.Failed)
          throw new WebException(r.ResultText, null, WebExceptionStatus.RequestCanceled, null);
      }

      result = Authenticate(session, profile);

      if (result == null) {
        throw new ImapException("appropriate authentication mechanism not found");
      }
      else if (result.Failed) {
        try {
          try {
            session.Disconnect(false);
          }
          catch (ImapConnectionException) {
            // ignore
          }
        }
        finally {
          session = null;
        }
      }

      return result;
    }

    private static ImapTransactionResult Authenticate(ImapSession session, IImapSessionProfile profile)
    {
      var authority = profile.Authority;
      var username = ImapStyleUriParser.GetUser(authority);
      var authMechanism = ImapStyleUriParser.GetAuthType(authority);

      /*
       * http://tools.ietf.org/html/rfc5092
       * 3.2. IMAP User Name and Authentication Mechanism
       * 
       *    An authentication mechanism (as used by the IMAP AUTHENTICATE
       *    command) can be expressed by adding ";AUTH=<enc-auth-type>" to the
       *    end of the user name in an IMAP URL.  When such an <enc-auth-type> is
       *    indicated, the client SHOULD request appropriate credentials from
       *    that mechanism and use the "AUTHENTICATE" command instead of the
       *    "LOGIN" command.  If no user name is specified, one MUST be obtained
       *    from the mechanism or requested from the user/configuration as
       *    appropriate.
       *
       *    If a user name is included with no authentication mechanism, then
       *    ";AUTH=*" is assumed.
       */

      // TODO: URLAUTH
      /*
      if (!session.ServerCapabilities.Has(ImapCapability.UrlAuth)
        throw new ImapWebException("URLAUTH incapable", WebExceptionStatus.ProtocolError);
      */
      var canFallback = false;

      if (authMechanism == null) {
        if (string.IsNullOrEmpty(username)) {
          authMechanism = ImapAuthenticationMechanism.Anonymous;
          canFallback = true;
        }
        else {
          authMechanism = ImapAuthenticationMechanism.SelectAppropriate;
        }
      }

      if (authMechanism == ImapAuthenticationMechanism.SelectAppropriate)
        return AuthenticateWithAppropriateMechanism(session,
                                                    profile.Credentials,
                                                    username,
                                                    GetUsingSaslMechanisms(profile.UsingSaslMechanisms ?? new string[0]));
      else if (authMechanism == ImapAuthenticationMechanism.Anonymous)
        return AuthenticateAsAnonymous(session,
                                       profile.AnonymousCredential,
                                       canFallback);
      else
        return AuthenticateWithSuppliedMechanism(session,
                                                 profile.Credentials,
                                                 username,
                                                 authMechanism);
    }

    private static IEnumerable<ImapAuthenticationMechanism> GetUsingSaslMechanisms(string[] usingSaslMechanisms)
    {
      var availableMechanisms = SaslClientMechanism.GetAvailableMechanisms();

      foreach (var mechanism in usingSaslMechanisms) {
        if (availableMechanisms.Exists(delegate(string val) {
          return string.Equals(val, mechanism, StringComparison.OrdinalIgnoreCase);
        }))
          yield return ImapAuthenticationMechanism.GetKnownOrCreate(mechanism);
      }
    }

    private static ImapTransactionResult AuthenticateWithAppropriateMechanism(ImapSession session, ICredentialsByHost credentials, string username, IEnumerable<ImapAuthenticationMechanism> usingSaslMechanisms)
    {
      ImapTransactionResult result = null;

      foreach (var mechanism in usingSaslMechanisms) {
        if (session.ServerCapabilities.IsCapable(mechanism)) {
          result = session.Authenticate(credentials, username, mechanism, true);

          if (result.Succeeded)
            break;
        }
      }

      if ((result == null || result.Failed) &&
          !session.ServerCapabilities.Has(ImapCapability.LoginDisabled))
        result = session.Login(credentials, username, true);

      return result;
    }

    private static ImapTransactionResult AuthenticateWithSuppliedMechanism(ImapSession session, ICredentialsByHost credentials, string username, ImapAuthenticationMechanism authMechanism)
    {
      /*
       * http://tools.ietf.org/html/rfc5092
       * 3.2. IMAP User Name and Authentication Mechanism
       * 
       *    An authentication mechanism (as used by the IMAP AUTHENTICATE
       *    command) can be expressed by adding ";AUTH=<enc-auth-type>" to the
       *    end of the user name in an IMAP URL.  When such an <enc-auth-type> is
       *    indicated, the client SHOULD request appropriate credentials from
       *    that mechanism and use the "AUTHENTICATE" command instead of the
       *    "LOGIN" command.  If no user name is specified, one MUST be obtained
       *    from the mechanism or requested from the user/configuration as
       *    appropriate.
       */
      return session.Authenticate(credentials, username, authMechanism, true);
    }

    private static ImapTransactionResult AuthenticateAsAnonymous(ImapSession session, NetworkCredential anonymousCredential, bool canFallback)
    {
      /*
       * http://tools.ietf.org/html/rfc5092
       * 3.2. IMAP User Name and Authentication Mechanism
       * 
       *    If no user name and no authentication mechanism are supplied, the
       *    client MUST authenticate as anonymous to the server.  If the server
       *    advertises AUTH=ANONYMOUS IMAP capability, the client MUST use the
       *    AUTHENTICATE command with ANONYMOUS [ANONYMOUS] SASL mechanism.  If
       *    SASL ANONYMOUS is not available, the (case-insensitive) user name
       *    "anonymous" is used with the "LOGIN" command and the Internet email
       *    address of the end user accessing the resource is supplied as the
       *    password.  The latter option is given in order to provide for
       *    interoperability with deployed servers.
       * 
       *    Note that, as described in RFC 3501, the "LOGIN" command MUST NOT be
       *    used when the IMAP server advertises the LOGINDISABLED capability.
       */
      ImapTransactionResult result = null;

      if (session.ServerCapabilities.IsCapable(ImapAuthenticationMechanism.Anonymous))
        // try AUTHENTICATE ANONYUMOUS
        result = session.Authenticate(anonymousCredential, null, ImapAuthenticationMechanism.Anonymous, true);

      if ((result == null || (result.Failed && canFallback)) &&
          !session.ServerCapabilities.Has(ImapCapability.LoginDisabled))
        // try anonymous LOGIN
        result = session.Login(anonymousCredential, null, true);

      return result;
    }
  }
}
